from recruiter_app.utilities.helpers import *


class RatingRecipe:
    def __init__(self, skills_required, data_skills_obj,
                 so_extra_score, primary_data_source, gh_extra_score=None,
                 so_dampen_factor=0.25, gh_dampen_factor=0.25):
        self.skills_required = skills_required
        self.data_skills_obj = data_skills_obj
        self.so_extra_score = so_extra_score
        self.gh_extra_score = gh_extra_score
        self.primary_data_source = primary_data_source
        self.so_dampen_factor = so_dampen_factor
        self.gh_dampen_factor = gh_dampen_factor
        self.result_queue_params = []

    def get_dampening_weight(self, skill_obj):
        """
        Get the regularization params for every feature
        :param skill_obj:
        :return:
        """
        dampening_weights = {}
        if skill_obj:
            scores_dict = get_all_feature_scores(skill_obj)
            if scores_dict["so_q_score"] and scores_dict["so_a_score"] and \
               scores_dict["so_q_score"] > 0 and scores_dict["so_a_score"] > 0:
                if scores_dict["so_q_score"] > scores_dict["so_a_score"]:
                    dampening_weights['dampen_so_question'] = self.so_dampen_factor
                if scores_dict["so_top_q_score"] > scores_dict["so_top_a_score"]:
                    dampening_weights['dampen_so_question_top'] = self.so_dampen_factor
            if scores_dict["gh_watchers_count"] and scores_dict["gh_forks_count"] and scores_dict["gh_stars_count"]:
                if sum([scores_dict["gh_watchers_count"],
                        scores_dict["gh_forks_count"]]) > scores_dict["gh_stars_count"]:
                    dampening_weights['dampen_gh_features'] = self.gh_dampen_factor
        return dampening_weights

    def calculate_rating(self, result_queue):
        """
        Main function for calculating the final score of the candidate
        :param result_queue: the queue to which this function pushes:
         - candidate id
         - job title for which the rating has happened
         - final score
        :return:
        """
        final_rating = 0
        for skill, skill_weight in self.skills_required.items():
            candidate_skills_dict = self.data_skills_obj.get(skill)
            if not candidate_skills_dict:
                continue
            dampening_weights = self.get_dampening_weight(candidate_skills_dict)
            so_question_weight = 1
            so_question_top_weight = 1
            gh_features_weight = 1
            if dampening_weights:
                dampen_so_question = dampening_weights.get("dampen_so_question")
                dampen_so_question_top = dampening_weights.get("dampen_so_question_top")
                dampen_gh_features = dampening_weights.get("dampen_gh_features")
                so_question_weight = 1 if not dampen_so_question else dampen_so_question
                so_question_top_weight = 1 if not dampen_so_question_top else dampen_so_question_top
                gh_features_weight = 1 if not dampen_gh_features else dampen_gh_features
            ft_scores_dict = get_all_feature_scores(candidate_skills_dict)
            skill_rating = sum([ft_scores_dict["so_q_score"] * so_question_weight,
                                ft_scores_dict["so_a_score"],
                                ft_scores_dict["so_top_q_score"] * so_question_top_weight,
                                ft_scores_dict["so_top_a_score"],
                                ft_scores_dict["so_top_a_isa_score"],
                                ft_scores_dict["gh_forks_count"] * gh_features_weight,
                                ft_scores_dict["gh_stars_count"],
                                ft_scores_dict["gh_watchers_count"] * gh_features_weight,
                                ft_scores_dict["gh_repo_count"]])
            final_rating += skill_rating * skill_weight
        if final_rating > 0:
            final_rating += self.so_extra_score
            if self.gh_extra_score:
                final_rating += self.gh_extra_score
        self.result_queue_params.extend([final_rating])
        result_queue.put(self.result_queue_params)



import mongoengine


class EvalCandidates(mongoengine.Document):
    meta = {
        "db_alias": 'core',
        "collection": 'evaluate_candidates'
    }
    candidate_id = mongoengine.IntField(required=True)
    jd_name = mongoengine.StringField(required=True)
    token = mongoengine.StringField(required=True)
    answer_1 = mongoengine.StringField()
    answer_2 = mongoengine.StringField()
    answer_3 = mongoengine.StringField()



import mongoengine


class GithubUser(mongoengine.Document):
    meta = {
        "db_alias": 'core',
        "collection": 'gh_users_details'
    }
    login_name = mongoengine.StringField(required=True, unique=True)
    name = mongoengine.StringField()
    email = mongoengine.StringField(default=None)
    bio = mongoengine.StringField(default=None)
    followers = mongoengine.IntField(default=0)
    public_repos = mongoengine.IntField(default=0)
    public_gists = mongoengine.IntField(default=0)
    for_jd = mongoengine.StringField(required=True)
    lang_score = mongoengine.DictField()
    alter_ego = mongoengine.DictField()


import mongoengine


class RedditUserProps(mongoengine.Document):
    meta = {
        "db_alias": 'core',
        "collection": 'reddit_user_props'
    }
    user_name = mongoengine.StringField(required=True, unique=True)
    for_jd = mongoengine.StringField(required=True)
    thread_props = mongoengine.DictField()
    alter_ego = mongoengine.DictField()



from recruiter_app.elt_app.service.rating_recipes import *
from recruiter_app.elt_app.service.mongo import *
from recruiter_app.elt_app.service.rating_prep import *
from recruiter_app.elt_app.service.merge_data import *
from recruiter_app.elt_app.service.exec_recipe import *



import multiprocessing as mp
from recruiter_app.utilities.helpers import create_batch
from recruiter_app.elt_app.service.mongo import service_create_candidate_rating
import threading


class ExecuteRecipe:
    def __init__(self):
        self.cpu_count = mp.cpu_count()
        self.result_queue = mp.Queue()

    def check_data_in_queue(self):
        """
        Keep checking the queue, if the is any rating data then push to database,
        if exit condition (FINISHED) is received then stop the thread.
        :return:
        """
        print("service_create_candidate_rating thread is starting")
        while True:
            queue_data = self.result_queue.get()
            if queue_data == "FINISHED":
                break
            if type(queue_data[2]) in [float, int]:
                service_create_candidate_rating(queue_data[0],
                                                queue_data[1],
                                                queue_data[2])
        print("service_create_candidate_rating thread is stopped")

    def start_processes(self, recipe_list):
        """
        starts two threads :
        1. for executing the recipes with multiprocessing
        2. for database update via queue mechanism
        :param recipe_list: list of class object of the recipes
        :return:
        """
        batch_process_th = threading.Thread(target=self.execute,
                                            args=(recipe_list,))
        save_to_db_th = threading.Thread(target=self.check_data_in_queue)
        save_to_db_th.start()
        batch_process_th.start()
        for th in [batch_process_th, save_to_db_th]:
            th.join()
        print("Successfully processed a batch of candidates")

    def execute(self, data_list):
        """
        Executes the recipes with multiprocessing
        :param data_list:
        :return:
        """
        for data_batch in create_batch(data_list, batch_size=self.cpu_count,
                                       description="\nProcessing sub-batch {0}"):
            process_list = []
            for data in data_batch:
                if len(data) != 3:
                    raise Exception(f"Data - {data} is not sufficient to start the process")
                p = mp.Process(target=data[2].calculate_rating,
                               args=(self.result_queue,))
                p.start()
                process_list.append(p)
            [process.join() for process in process_list]
        self.result_queue.put("FINISHED")





"""
Throttle:
- 30 req per sec
- 10k req daily
- limit -> pagesize:100
- limit -> id size : 100

"""
import requests
import os
from recruiter_app.utilities import helpers
import time
import urllib.parse
import asyncio
import aiohttp
from datetime import datetime


class StackExchangeData:
    quota_remaining = 0
    in_backoff_state = False
    back_off_till = None

    def __init__(self):
        config_props = helpers.get_prop("DATA_SOURCES").get("STACKOVERFLOW")
        self.access_token = os.environ['access_token']
        self.client_id = os.environ['client_id']
        self.session = requests.sessions.Session()
        self.default_params = {
            'order': 'desc',
            'sort': 'reputation',
            'site': 'stackoverflow',
            'client_id': self.client_id,
            'key': self.access_token
        }
        self.host = config_props.get('HOST')
        self.top_tag_url = config_props.get('user_top_tags')
        self.get_users_url = config_props.get('users')
        self.get_users_top_q = config_props.get('user_top_questions')
        self.get_users_top_a = config_props.get('user_top_answers')
        self.task_get_users = []

    def _fetch(self, url, **kwargs):
        """
        Synchronous http request call
        :param url: url string
        :param kwargs: params in url
        :return: response object
        """
        time.sleep(2)
        url = "".join([self.host, url])
        content = self.session.get(url, **kwargs)
        if content.status_code != 200:
            raise Exception(f"HTTP request {url} failed, err code : {content.status_code} \n"
                            f"Message : {content.content}")
        content = content.json()
        print(f"{'*' * 10}Quota remaining: {content.get('quota_remaining')}{'*' * 10}")
        if content.get('quota_remaining') <= 100:
            raise Exception("WARNING Limit is 100!!!!")
        if content.get('backoff'):
            print(f"backoff parameter received must wait for {content.get('backoff')}s")
            time.sleep(int(content.get('backoff')))
        return content

    async def handle_backoff(self, backoff):
        if StackExchangeData.quota_remaining < 1:
            print("stackexchange api Quota has been exhausted, will now wait till the reset")
            utc_midnight = datetime(2022, 9, datetime.utcnow().day + 1, 0, 0, 0, 0)
            utc_now = datetime.utcnow()
            backoff = utc_midnight.timestamp() - utc_now.timestamp()
        backoff = int(backoff)
        StackExchangeData.in_backoff_state = True
        print(f"backoff parameter received must wait for {backoff}s")
        StackExchangeData.back_off_till = time.time() + backoff + 2
        await asyncio.sleep(backoff + 2)
        print("backoff period is over resuming execution")
        StackExchangeData.in_backoff_state = False

    async def _async_fetch(self, url, **kwargs):
        """
        Async task to send the http requests
        :param url: complete url
        :param kwargs: keyword argument for passing param in http requests
        :return: "items" values in response received from the http requests
        """
        if StackExchangeData.in_backoff_state and time.time() < StackExchangeData.back_off_till:
            wait_period = StackExchangeData.back_off_till - time.time()
            if wait_period > 0:
                print(f"In backoff state and wait period is {wait_period}")
                await asyncio.sleep(wait_period)
            print("backoff period is over resuming execution")
        async with aiohttp.ClientSession() as session:
            async with session.get(url, **kwargs) as response:
                if response.status != 200:
                    if response.status == 403:
                        print(f"Cannot access this profile {url}")
                    else:
                        raise Exception(f"HTTP request {url} failed, err code : {response.status} \n"
                                        f"Message : {response.content}")
                content = await response.json()
                StackExchangeData.quota_remaining = content.get('quota_remaining')
                print(f"{'*' * 10}Quota remaining: {StackExchangeData.quota_remaining}{'*' * 10}")
                backoff = content.get('backoff')
                if backoff or StackExchangeData.quota_remaining < 1:
                    await self.handle_backoff(backoff)
                return content.get('items')

    def get_top_qna(self, query_type, user_id, tag_name, page=1, pagesize=100):
        """
        Fetch top qna data for a user for a tag - tag_name
        :param query_type: question tag or answer tag
        :param user_id: id of the user
        :param tag_name: name of tag
        :param page: page number
        :param pagesize: number of results per page
        :return: response object from the request
        """
        params = self.default_params.copy()
        params.update({"page": str(page),
                       "pagesize": str(pagesize),
                       "sort": "activity"})
        if query_type == 'question':
            url = self.get_users_top_q.format(user_id,
                                              urllib.parse.quote(tag_name))
        else:
            url = self.get_users_top_a.format(user_id,
                                              urllib.parse.quote(tag_name))
        response_obj = self._fetch(url, params=params)
        response_obj = response_obj.get('items')
        return response_obj

    def get_tags(self, users_list, page=1, pagesize=100):
        """
        Fetch tags of every user
        :param users_list: list of user ids (max - 100)
        :param page: page number
        :param pagesize: number of result per page (max 100)
        :yield: back response object and has_more indicator which tells
        the called function whether to continue or stop
        """
        users_list = [str(user.user_id) for user in users_list]
        print(f"Getting tags from stackoverflow for {len(users_list)} users")
        if len(users_list) > 100:
            raise Exception(f"Trying to fetch details for {len(users_list)} which is > 100")
        users_list = ";".join(users_list)
        params = self.default_params.copy()
        params.update({"page": str(page), "pagesize": str(pagesize)})
        url = self.top_tag_url.format(users_list)
        response_obj = self._fetch(url, params=params)
        has_more = response_obj.get('has_more')
        response_obj = response_obj.get('items')
        yield response_obj, has_more

    async def get_users(self, page_counts, page_size=100):
        """
        Calls users/ api for fetching 100 users per page,
        100 pages are requested at a time using helpers.create_batch
        and executed in asynch mode
        # TODO : Open issue : # 10
        :param page_counts: number of pages
        :param page_size: number of results returned per page
        :return:
        """
        users_obj_list = []
        url = "".join([self.host, self.get_users_url])
        for page_batch in helpers.create_batch(page_counts, start=1):
            for page in page_batch:
                print(f"Creating task for page {page}")
                params = self.default_params.copy()
                params.update({"page": str(page), "pagesize": str(page_size)})
                self.task_get_users.append(asyncio.create_task(self._async_fetch(url,
                                                                                 params=params)))
            users_obj_per_pages = await asyncio.gather(*self.task_get_users)
            [[users_obj_list.append(user_obj) for user_obj in user_obj_page]
             for user_obj_page in users_obj_per_pages]
            self.task_get_users = []
        return users_obj_list



from recruiter_app.elt_app.collections.gh_users_details import GithubUser
from recruiter_app.elt_app.collections.reddit_user_props import RedditUserProps
from recruiter_app.elt_app.collections.so_users_details import StackOverflowUser
from recruiter_app.elt_app.collections.candidate_ratings import CandidateRatings
import re


def service_create_record(user_data):
    """
    Updates the data received from the api call
    into so_users_details collection
    :param user_data: dict containing the basic
    details of user
    :return: _id in mongo db
    """
    if not StackOverflowUser.objects().filter(user_id=user_data.get('user_id')):
        user_obj = {'user_id': user_data.get('user_id'),
                    'reputation': user_data.get('reputation'),
                    'accept_rate': user_data.get('accept_rate'),
                    'display_name': user_data.get('display_name'),
                    'gold_badge_count': user_data.get('badge_counts').get('gold'),
                    'silver_badge_count': user_data.get('badge_counts').get('silver'),
                    'bronze_badge_count': user_data.get('badge_counts').get('bronze'),
                    'website_url': user_data.get('website_url')}
        so_data = StackOverflowUser()
        so_data.display_name = user_obj['display_name']
        so_data.user_id = user_obj['user_id']
        so_data.gold_badge_count = user_obj['gold_badge_count']
        so_data.silver_badge_count = user_obj['silver_badge_count']
        so_data.bronze_badge_count = user_obj['bronze_badge_count']
        so_data.accept_rate = user_obj['accept_rate']
        so_data.reputation = user_obj['reputation']
        so_data.website_url = user_obj['website_url']

        so_data.save()
        return so_data.id
    print(f"Stackoverflow User {user_data.get('user_id')} already present in db")


def service_update_tags_data_available(has_more, user_obj, threshold):
    user_tags = StackOverflowUser.objects().filter(user_id=user_obj.user_id)[0]
    if len(user_tags.tags_score) > threshold or (len(user_tags.tags_score) < threshold and not has_more):
        user_obj.tags_data_available = True
        user_obj.save()
        reset_page = True
    else:
        reset_page = False
    return reset_page


def service_update_top_qna(user, data_obj_list, tag_name, query_type):
    """
    Update top question and answer data for the tag - tag_name
    :param user: id of the user
    :param data_obj_list: top qna data list
    :param tag_name: name of tag
    :param query_type: type of tag - question or answer
    :return: None
    """
    tag_name = f"tag_{tag_name}"
    tag_name = re.sub("[`~!@#$%^&*()\\]\\[+={}/|:;\"\'<>,.?-]", '_', tag_name)
    user_tags_score_obj = StackOverflowUser.objects(user_id=user)[0]
    for data_obj in data_obj_list:
        if query_type == 'question':
            user_tags_score_obj.tags_score[tag_name][f"top_question_{data_obj.get('question_id')}_score"] = \
                data_obj.get('score')
        else:
            user_tags_score_obj.tags_score[tag_name][f"top_answer_{data_obj.get('answer_id')}_score"] = \
                data_obj.get('score')
            user_tags_score_obj.tags_score[tag_name][f"top_answer_{data_obj.get('answer_id')}_is_accepted"] = \
                data_obj.get('is_accepted')
    user_tags_score_obj.tags_top_qna_data_available = True
    user_tags_score_obj.save()


def service_create_user_tags_records(data):
    """
    Updates the tags_score dict field with format
    {
        "tags_score": {
            "tag_name": {
                {"question_score": {question_score},
                 "answer_score": {answer_score},
                 "name": {original_tag_name}
            }
        }
    }
    :param data:
    :return: None
    """
    print("Creating the user tag records for new users")
    for user_tag_data in data:
        user_data = StackOverflowUser.objects().filter(user_id=user_tag_data.get('user_id'))[0]
        tag_obj_name = f"tag_{user_tag_data.get('tag_name')}"
        tag_obj_name = re.sub("[`~!@#$%^&*()\\]\\[+={}/|:;\"\'<>,.?-]", '_', tag_obj_name)
        user_data.tags_score[tag_obj_name] = {}
        if user_tag_data.get('question_count') > 0:
            score_per_q = user_tag_data.get('question_score') / user_tag_data.get('question_count')
            user_data.tags_score[tag_obj_name]['question_score'] = score_per_q
        if user_tag_data.get('answer_count') > 0:
            score_per_a = user_tag_data.get('answer_score') / user_tag_data.get('answer_count')
            user_data.tags_score[tag_obj_name]['answer_score'] = score_per_a
        user_data.tags_score[tag_obj_name]['name'] = user_tag_data.get('tag_name')
        user_data.tags_data_available = True
        user_data.save()


def service_create_github_users(jd_name, data_obj, repo_objs):
    """
    Creating the collection with user details and user repo details
    :param jd_name: job title
    :param data_obj: user details dict
    :param repo_objs: user repo details dict
    :return: None
    """
    gh_user = GithubUser()
    gh_user.login_name = data_obj.get('login')
    gh_user.name = data_obj.get('name') if data_obj.get('name') else data_obj.get('login')
    gh_user.email = data_obj.get('email')
    gh_user.bio = data_obj.get('bio')
    gh_user.followers = data_obj.get('followers')
    gh_user.public_repos = data_obj.get('public_repos')
    gh_user.public_gists = data_obj.get('public_gists')
    gh_user.for_jd = jd_name
    gh_user.lang_score = {}
    for idx, repo_obj in enumerate(repo_objs, start=1):
        prog_language = repo_obj.get('language')
        if prog_language:
            prog_language = prog_language.replace(' ', '_').lower()
            if f"lang_{prog_language}" not in gh_user.lang_score:
                gh_user.lang_score[f"lang_{prog_language}"] = {}

            gh_user.lang_score[f"lang_{prog_language}"] \
                [f"repo_{idx}_forks"] = repo_obj.get('forks_count')

            gh_user.lang_score[f"lang_{prog_language}"] \
                [f"repo_{idx}_stars"] = repo_obj.get('stargazers_count')

            gh_user.lang_score[f"lang_{prog_language}"] \
                [f"repo_{idx}_watchers"] = repo_obj.get('watchers_count')
    gh_user.save()


def service_create_rd_record(users_obj, jd):
    """
    Create reddit users collections
    :param users_obj: users dict
    :param jd: name of job title
    :return: None
    """
    print(f"\nInserting users_obj for jd - {jd} to collection")
    for user_props in users_obj:
        if RedditUserProps.objects(user_name=user_props):
            rup = RedditUserProps.objects(user_name=user_props)[0]
            rup.update(thread_props=users_obj[user_props])
        else:
            if user_props and jd:
                rup = RedditUserProps()
                rup.user_name = user_props
                rup.for_jd = jd
                rup.thread_props = users_obj[user_props]
                rup.save()


def service_update_alter_ego(source_obj, value, label):
    """
    updates the alter_ego field in every user collection
    in source_obj with value with label
    :param source_obj: DOM object for which data is updated
    :param value: id of user in another source
    :param label: name of the source
    :return:
    """
    if source_obj.alter_ego:
        # incase there is already an
        # existing alter_ego in another source
        so_existing_obj = source_obj.alter_ego
        so_existing_obj.update({label: f"{value}"})
        source_obj.update(alter_ego=so_existing_obj)
    else:
        source_obj.update(alter_ego={label: f"{value}"})


def service_create_candidate_rating(user_id, jd_name, rating):
    cd_rating_obj = CandidateRatings()
    cd_rating_obj.user_id = user_id
    cd_rating_obj.jd_name = jd_name
    cd_rating_obj.rating = rating
    cd_rating_obj.save()



import sys
import asyncio
import requests
from recruiter_app.elt_app import *
from recruiter_app.utilities import *


def start(mode):
    """
    Main function which is called for executing the recruiter_app
    :param mode: parameter which will decide the mode of execution
    possible values are -
    extract_stackoverflow,
    extract_github,
    extract_reddit,
    merge_data,
    rating_prep,
    rate_candidate,
    get_top_n_candidates
    :return:
    """
    if mode == "extract_stackoverflow":
        if sys.platform.startswith("win"):
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        print("Starting Stackoverflow factory")
        so_factory = SOFactory('cloud')
        so_factory.create_user_collection()
        print("Stackoverflow data extracted successfully")

    elif mode == "extract_github":
        print("Starting Github factory")
        gh_factory = GHFactory('cloud')
        jds = ['web developer', 'python developer', 'data scientist',
               "full stack developer", "devops engineer"]
        for jd in jds:
            print(f"Now mining for jd - {jd}")
            gh_factory.build_users_collection(jd)
            print(f"Mining target achieved for {jd}, moving to next jd")
        print("Github data extracted successfully")

    elif mode == "extract_reddit":
        print("Starting Reddit factory")
        all_jds = ["web_dev_thread", "devops_thread",
                   "cyber_seq_thread", "ml_thread", "py_thread"]
        rf = RedditFactory('cloud')
        rf.build_users_collection(all_jds)
        print("Reddit data extracted successfully")

    elif mode == "merge_data":
        print("Starting to merge data from different sources")
        mds = MergeDataSources()
        mds.merge_data()
        print("Successfully completed merge data")

    elif mode == "rating_prep":
        print("Creating csv file for labelling the tags/langs")
        so_user_list = StackOverflowUser.objects()
        gh_user_list = GithubUser.objects()
        run_prep_skill_count_csv(so_user_list, gh_user_list)
        print("Successfully completed rating_prep")

    elif mode == "rate_candidate":
        jd_list = ["python developer", "data scientist", "web developer", "devops engineer"]
        for jd in jd_list:
            print(f"rating for job description : {jd}")
            re_obj = RatingEngine(jd)
            re_obj.start()
            print(f"finished rating for jd : {jd}")

    elif mode == "get_top_n_candidates":
        mongo_global_init(host_type="cloud")
        n = int(os.environ["top_n"])
        jd_list = ["python developer", "data scientist", "web developer", "devops engineer"]
        for jd in jd_list:
            print(f"for jd : {jd}")
            top_candidates = CandidateRatings.objects(jd_name=jd).order_by('-rating').limit(n)
            for candidate in top_candidates:
                print(f"Sending mail to {candidate.user_id}with rating f{candidate.rating}")
                mail_data_payload = {"candidate_id": str(candidate.user_id),
                                     "jd_name": jd}
                req = requests.post("http://127.0.0.1:5000/notify_candidates",
                                    json=mail_data_payload)
                print(f"request sent status is : {req.status_code}")

    else:
        raise Exception(f"mode {mode} is not recognized, please enter any of: "
                        f"stackoverflow, github, redit")


if __name__ == '__main__':
    print("Welcome to hush hush recruiter cmd line tool")
    if len(sys.argv) < 2:
        raise Exception("Enter the mode of execution")
    mode = sys.argv[1]
    print(f"Tool will run in {mode} mode")
    start(mode)
    print("Execution completed successfully")



import mongoengine


class CandidateRatings(mongoengine.Document):
    meta = {
        "db_alias": 'core',
        "collection": 'candidate_ratings'
    }
    jd_name = mongoengine.StringField(required=True)
    user_id = mongoengine.IntField(required=True)
    rating = mongoengine.FloatField(default=0.0)
